from __future__ import division

from math import ceil

from django.contrib import messages
from django.contrib.admin.views.decorators import staff_member_required
from django.http import Http404
from django.shortcuts import redirect, render

from redis.exceptions import ResponseError
from rq import requeue_job, Worker
from rq.exceptions import NoSuchJobError
from rq.job import Job
from rq.registry import (DeferredJobRegistry, FinishedJobRegistry,
                         StartedJobRegistry)

from django_rq.queues import get_connection, get_queue_by_index
from django_rq.settings import QUEUES_LIST
from . import settings

from .serializer import default_secure_serializer as secure_serializer


def use_actual_name(job):
    if 'secure_redis.secure_rq.secure_job_proxy' != job.func_name:
        return
    encrypted_func_name = job.args[0]
    actual_func_name = secure_serializer.loads(encrypted_func_name)
    job.func_name = actual_func_name


@staff_member_required
def stats(request):
    queues = []
    for index, config in enumerate(QUEUES_LIST):

        queue = get_queue_by_index(index)
        connection = queue.connection

        queue_data = {
            'name': queue.name,
            'jobs': queue.count,
            'index': index,
            'connection_kwargs': connection.connection_pool.connection_kwargs
        }

        if queue.name == 'failed':
            queue_data['workers'] = '-'
            queue_data['finished_jobs'] = '-'
            queue_data['started_jobs'] = '-'
            queue_data['deferred_jobs'] = '-'

        else:
            connection = get_connection(queue.name)
            all_workers = Worker.all(connection=connection)
            queue_workers = [worker for worker in all_workers if queue in worker.queues]
            queue_data['workers'] = len(queue_workers)

            finished_job_registry = FinishedJobRegistry(queue.name, connection)
            started_job_registry = StartedJobRegistry(queue.name, connection)
            deferred_job_registry = DeferredJobRegistry(queue.name, connection)
            queue_data['finished_jobs'] = len(finished_job_registry)
            queue_data['started_jobs'] = len(started_job_registry)
            queue_data['deferred_jobs'] = len(deferred_job_registry)

        queues.append(queue_data)

    context_data = {'queues': queues}
    return render(request, 'django_rq/stats.html', context_data)


@staff_member_required
def jobs(request, queue_index):
    queue_index = int(queue_index)
    queue = get_queue_by_index(queue_index)

    items_per_page = 100
    num_jobs = queue.count
    page = int(request.GET.get('page', 1))

    if num_jobs > 0:
        last_page = int(ceil(num_jobs / items_per_page))
        page_range = range(1, last_page + 1)
        offset = items_per_page * (page - 1)
        jobs = queue.get_jobs(offset, items_per_page)
    else:
        jobs = []
        page_range = []

    ###
    # Custom logic here
    for job in jobs:
        use_actual_name(job)
    ##

    context_data = {
        'queue': queue,
        'queue_index': queue_index,
        'jobs': jobs,
        'num_jobs': num_jobs,
        'page': page,
        'page_range': page_range,
        'job_status': 'Queued',
    }
    return render(request, 'django_rq/jobs.html', context_data)


@staff_member_required
def finished_jobs(request, queue_index):
    queue_index = int(queue_index)
    queue = get_queue_by_index(queue_index)

    registry = FinishedJobRegistry(queue.name, queue.connection)

    items_per_page = 100
    num_jobs = len(registry)
    page = int(request.GET.get('page', 1))
    jobs = []

    if num_jobs > 0:
        last_page = int(ceil(num_jobs / items_per_page))
        page_range = range(1, last_page + 1)
        offset = items_per_page * (page - 1)
        job_ids = registry.get_job_ids(offset, items_per_page)

        for job_id in job_ids:
            try:
                jobs.append(Job.fetch(job_id, connection=queue.connection))
            except NoSuchJobError:
                pass

    else:
        page_range = []

    ###
    # Custom logic here
    for job in jobs:
        use_actual_name(job)
    ##

    context_data = {
        'queue': queue,
        'queue_index': queue_index,
        'jobs': jobs,
        'num_jobs': num_jobs,
        'page': page,
        'page_range': page_range,
        'job_status': 'Finished',
    }
    return render(request, 'django_rq/jobs.html', context_data)


@staff_member_required
def started_jobs(request, queue_index):
    queue_index = int(queue_index)
    queue = get_queue_by_index(queue_index)

    registry = StartedJobRegistry(queue.name, queue.connection)

    items_per_page = 100
    num_jobs = len(registry)
    page = int(request.GET.get('page', 1))
    jobs = []

    if num_jobs > 0:
        last_page = int(ceil(num_jobs / items_per_page))
        page_range = range(1, last_page + 1)
        offset = items_per_page * (page - 1)
        job_ids = registry.get_job_ids(offset, items_per_page)

        for job_id in job_ids:
            try:
                jobs.append(Job.fetch(job_id, connection=queue.connection))
            except NoSuchJobError:
                pass

    else:
        page_range = []

    ###
    # Custom logic here
    for job in jobs:
        use_actual_name(job)
    ##

    context_data = {
        'queue': queue,
        'queue_index': queue_index,
        'jobs': jobs,
        'num_jobs': num_jobs,
        'page': page,
        'page_range': page_range,
        'job_status': 'Started',
    }
    return render(request, 'django_rq/jobs.html', context_data)


@staff_member_required
def deferred_jobs(request, queue_index):
    queue_index = int(queue_index)
    queue = get_queue_by_index(queue_index)

    registry = DeferredJobRegistry(queue.name, queue.connection)

    items_per_page = 100
    num_jobs = len(registry)
    page = int(request.GET.get('page', 1))
    jobs = []

    if num_jobs > 0:
        last_page = int(ceil(num_jobs / items_per_page))
        page_range = range(1, last_page + 1)
        offset = items_per_page * (page - 1)
        job_ids = registry.get_job_ids(offset, items_per_page)

        for job_id in job_ids:
            try:
                jobs.append(Job.fetch(job_id, connection=queue.connection))
            except NoSuchJobError:
                pass

    else:
        page_range = []

    ###
    # Custom logic here
    for job in jobs:
        use_actual_name(job)
    ##

    context_data = {
        'queue': queue,
        'queue_index': queue_index,
        'jobs': jobs,
        'num_jobs': num_jobs,
        'page': page,
        'page_range': page_range,
        'job_status': 'Deferred',
    }
    return render(request, 'django_rq/jobs.html', context_data)


@staff_member_required
def job_detail(request, queue_index, job_id):
    queue_index = int(queue_index)
    queue = get_queue_by_index(queue_index)
    try:
        job = Job.fetch(job_id, connection=queue.connection)
    except NoSuchJobError:
        raise Http404("Couldn't find job with this ID: %s" % job_id)

    ###
    # Custom logic here
    use_actual_name(job)
    ##

    context_data = {
        'queue_index': queue_index,
        'job': job,
        'queue': queue,
    }
    return render(request, 'django_rq/job_detail.html', context_data)


@staff_member_required
def delete_job(request, queue_index, job_id):
    queue_index = int(queue_index)
    queue = get_queue_by_index(queue_index)
    job = Job.fetch(job_id, connection=queue.connection)

    if request.method == 'POST':
        # Remove job id from queue and delete the actual job
        queue.connection._lrem(queue.key, 0, job.id)
        job.delete()
        messages.info(request, 'You have successfully deleted %s' % job.id)
        return redirect('rq_jobs', queue_index)

    ###
    # Custom logic here
    use_actual_name(job)
    ##

    context_data = {
        'queue_index': queue_index,
        'job': job,
        'queue': queue,
    }
    return render(request, 'django_rq/delete_job.html', context_data)


@staff_member_required
def requeue_job_view(request, queue_index, job_id):
    queue_index = int(queue_index)
    queue = get_queue_by_index(queue_index)
    job = Job.fetch(job_id, connection=queue.connection)

    if request.method == 'POST':
        requeue_job(job_id, connection=queue.connection)
        messages.info(request, 'You have successfully requeued %s' % job.id)
        return redirect('rq_job_detail', queue_index, job_id)

    ###
    # Custom logic here
    use_actual_name(job)
    ##

    context_data = {
        'queue_index': queue_index,
        'job': job,
        'queue': queue,
    }
    return render(request, 'django_rq/delete_job.html', context_data)


@staff_member_required
def clear_queue(request, queue_index):
    queue_index = int(queue_index)
    queue = get_queue_by_index(queue_index)

    if request.method == 'POST':
        try:
            queue.empty()
            messages.info(request, 'You have successfully cleared the queue %s' % queue.name)
        except ResponseError as e:
            if 'EVALSHA' in e.message:
                messages.error(request, 'This action is not supported on Redis versions < 2.6.0, please use the bulk delete command instead')
            else:
                raise e
        return redirect('rq_jobs', queue_index)

    context_data = {
        'queue_index': queue_index,
        'queue': queue,
    }
    return render(request, 'django_rq/clear_queue.html', context_data)


@staff_member_required
def requeue_all(request, queue_index):
    queue_index = int(queue_index)
    queue = get_queue_by_index(queue_index)
    jobs = queue.get_jobs()

    if request.method == 'POST':
        # Confirmation received
        for job in jobs:
            requeue_job(job.id, connection=queue.connection)

        messages.info(request, 'You have successfully requeued all %d jobs!' % len(jobs))
        return redirect('rq_jobs', queue_index)

    context_data = {
        'queue_index': queue_index,
        'queue': queue,
        'total_jobs':len(jobs),
    }

    return render(request, 'django_rq/requeue_all.html', context_data)


@staff_member_required
def actions(request, queue_index):
    queue_index = int(queue_index)
    queue = get_queue_by_index(queue_index)

    if request.method == 'POST' and request.POST.get('action', False):
        # confirm action
        if request.POST.get('_selected_action', False):
            context_data = {
                'queue_index': queue_index,
                'action': request.POST['action'],
                'job_ids': request.POST.getlist('_selected_action'),
                'queue': queue,
            }
            return render(request, 'django_rq/confirm_action.html', context_data)

        # do confirmed action
        elif request.POST.get('job_ids', False):
            job_ids = request.POST.getlist('job_ids')

            if request.POST['action'] == 'delete':
                for job_id in job_ids:
                    job = Job.fetch(job_id, connection=queue.connection)
                    # Remove job id from queue and delete the actual job
                    queue.connection._lrem(queue.key, 0, job.id)
                    job.delete()
                messages.info(request, 'You have successfully deleted %s jobs!' % len(job_ids))
            elif request.POST['action'] == 'requeue':
                for job_id in job_ids:
                    requeue_job(job_id, connection=queue.connection)
                messages.info(request, 'You have successfully requeued %d  jobs!' % len(job_ids))

    return redirect('rq_jobs', queue_index)



from __future__ import unicode_literals

import mock
import datetime

import django.test

import secure_redis.secure_rq


@secure_redis.secure_rq.job
def dummy(*args, **kwargs):
    pass


class SecureRedisRQTestCase(django.test.TestCase):
    @mock.patch('secure_redis.secure_rq.execute')
    @mock.patch('rq.job.Job.save')
    def test_delay_basic(self, save_method, execute_method):
        args = [1, 2, 3, ]
        kwargs = {'param1': 'A', 'param2': 'B', }
        dummy.delay(*args, **kwargs)

    @mock.patch('rq.queue.Queue.enqueue_call')
    def test_delay_no_plain_parameters(self, enqueue_call_method):
        args = [1, 2, 3, ]
        kwargs = {'param1': 'A', 'param2': 'B', }
        dummy.delay(*args, **kwargs)
        enqueue_call_method.assert_called_once()
        self.assertEqual('secure_job_proxy', enqueue_call_method.call_args[0][0].func_name)
        # Get the intersection between sent paramters and actual ones and ensure there are no intersection
        self.assertFalse([i for i in args if i in enqueue_call_method.call_args[1]['args']])
        self.assertFalse([i for i in kwargs if i in enqueue_call_method.call_args[1]['args']])
        self.assertFalse([i for i in kwargs if i in enqueue_call_method.call_args[1]['kwargs']])

    @mock.patch('secure_redis.secure_rq.execute')
    @mock.patch('rq.job.Job.save')
    def test_delay_parameters_recovered(self, save_method, execute_method):
        args = [1, 2, 3, ]
        kwargs = {'param1': 'A', 'param2': 'B', }
        dummy.delay(*args, **kwargs)
        execute_method.assert_called_with('secure_redis.tests.test_secure_rq.dummy', *args, **kwargs)

    @mock.patch('rq_scheduler.scheduler.Scheduler.enqueue_at')
    def test_enqueue_at_called(self, enqueue_at_method):
        args = [1,2, ]
        kwargs = {'a': '1', }
        t = datetime.datetime.now()
        dummy.enqueue_at(t, 'default', *args, **kwargs)
        enqueue_at_method.assert_called_once()
        self.assertIn(t, enqueue_at_method.call_args[0])

    @mock.patch('secure_redis.secure_rq.execute')
    @mock.patch('rq.job.Job.save')
    def test_enqueue_at_executed(self, save_method, execute_method):
        args = [1,2, ]
        kwargs = {'a': '1', }
        t = datetime.datetime.now()
        job = dummy.enqueue_at(t, 'default', *args, **kwargs)
        job.perform()
        execute_method.assert_called_with('secure_redis.tests.test_secure_rq.dummy', *args, **kwargs)

    @mock.patch('rq_scheduler.scheduler.Scheduler.schedule')
    def test_scheduled_once_called(self, schedule_method):
        interval = 60
        dummy.schedule_once(interval)
        schedule_method.assert_called_once()
        self.assertIn(interval, schedule_method.call_args[1].values())



from __future__ import unicode_literals

from cryptography.fernet import Fernet

import django_redis.serializers.pickle

from . import settings


class SecureSerializer(django_redis.serializers.pickle.PickleSerializer):
    def __init__(self, options):
        super(SecureSerializer, self).__init__(options)
        param_key = options.get("REDIS_SECRET_KEY")
        key = param_key.encode("utf-8")
        self.crypter = Fernet(key)

    def dumps(self, value):
        val = super(SecureSerializer, self).dumps(value)
        return self.crypter.encrypt(val)

    def loads(self, value):
        val = self.crypter.decrypt(value)
        return super(SecureSerializer, self).loads(val)


default_secure_serializer = SecureSerializer(settings.get_secure_cache_opts())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re
import os
from setuptools import setup, find_packages


with open('secure_redis/__init__.py', 'r') as init_file:
    version = re.search(
        '^__version__ = [\'"]([^\'"]+)[\'"]',
        init_file.read(),
        re.MULTILINE,
    ).group(1)

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-redis-secure',
    version=version,
    packages=find_packages(),
    license='MIT',
    include_package_data=True,
    description=(
        'Django caching plugin for django-redis that adds a Serializer class and configuration to support '
        'transparent, symmetrical encryption of cached values using the python cryptography library. This '
        'plugin also provides encryption for django-rq jobs by simply using the @secure_redis.secure_rq.job '
        'decorator to annotate the task method instead of using @django_rq.job.'
    ),
    url='http://github.com/foundertherapy/django-redis-secure/',
    download_url='https://github.com/foundertherapy/django-redis-secure/archive/' + version + '.tar.gz',
    author='FounderTherapy',
    author_email='oss@foundertherapy.com',
    install_requires=[
        'Django>=1.9',
        'cryptography>=0.8.2',
        'django-rq>=0.9.1',
        'rq>=0.6.0',
        'rq-scheduler>=0.6.1',
        'django-redis',
    ],
    keywords=['encryption', 'django', 'redis', 'rq', ],
)



from __future__ import (division, unicode_literals)
from functools import wraps
from collections import defaultdict
import logging
import datetime

from django.conf import settings as global_settings

import django_rq
from django_rq.queues import get_queue
import rq.utils
from rq.compat import string_types
from rq.defaults import DEFAULT_RESULT_TTL
from rq.queue import Queue

from .serializer import default_secure_serializer as secure_serializer


logger = logging.getLogger(__name__)


def execute(method_name, *args, **kwargs):
    return rq.utils.import_attribute(method_name)(*args, **kwargs)


def secure_job_proxy(*args, **kwargs):
    """
    The is a proxy method, each method wanted to be stored securely, it will be directed to this method. With first
    argument is the encrypted original method name.
    :param args: Always three parameters, first is the encrypted method name, second is the encrypted actual args,
    third is the encrypted actual kwargs
    :return: actual method return value
    """
    actual_function_name = secure_serializer.loads(args[0])
    decrypted_args = secure_serializer.loads(args[1])
    kwargs = secure_serializer.loads(args[2])
    return execute(actual_function_name, *decrypted_args, **kwargs)


def job(func_or_queue, connection=None, *args, **kwargs):
    class _rq_job(object):
        def __init__(self, queue, connection=None, timeout=None,
                     result_ttl=DEFAULT_RESULT_TTL, ttl=None):
            """A decorator that adds :
            ``delay`` method to the decorated function,
            which in turn creates a RQ job when called. Accepts a required
            ``queue`` argument that can be either a ``Queue`` instance or a string
            denoting the queue name.  For example:

                @job(queue='default')
                def simple_add(x, y):
                    return x + y

                simple_add.delay(1, 2) # Puts simple_add function into queue

            ``enqueue_at`` method, which creates RQ job to be executed in the specified argument date. ``enqueue_at``
            method take the ``target_date`` as first argument, ``scheduler_name`` as second argument optional argument,
            the later argument is used to specify which ``django_rq`` to be used to enqueue job for

            ``schedule_once`` method, schedule job once or reschedule when interval changes, if job already exists,
            this will not do anything.
            """
            self.queue = queue
            self.connection = connection
            self.timeout = timeout
            self.result_ttl = result_ttl
            self.ttl = ttl

        def __call__(self, f):
            @wraps(f)
            def delay(*args, **kwargs):
                if isinstance(self.queue, string_types):
                    queue = Queue(name=self.queue, connection=self.connection)
                else:
                    queue = self.queue
                depends_on = kwargs.pop('depends_on', None)
                function_name = '{}.{}'.format(f.__module__, f.__name__)
                encrypted_args = [secure_serializer.dumps(function_name)]
                encrypted_args += [secure_serializer.dumps(args)]

                encrypted_args += [secure_serializer.dumps(kwargs)]
                return queue.enqueue_call(secure_job_proxy, args=encrypted_args, kwargs={},
                                          timeout=self.timeout, result_ttl=self.result_ttl,
                                          ttl=self.ttl, depends_on=depends_on)

            @wraps(f)
            def enqueue_at(target_date, scheduler_name='default', *args, **kwargs):
                function_name = '{0}.{1}'.format(f.__module__, f.__name__)
                encrypted_args = [secure_serializer.dumps(function_name)]
                encrypted_args += [secure_serializer.dumps(args)]

                encrypted_args += [secure_serializer.dumps(kwargs)]

                scheduler = django_rq.get_scheduler(scheduler_name)
                proxy_method_name = '{}.{}'.format(secure_job_proxy.__module__, secure_job_proxy.__name__)
                return scheduler.enqueue_at(target_date, proxy_method_name, *encrypted_args, **{})

            @wraps(f)
            def schedule_once(interval, timeout=None):
                """
                Schedule job once or reschedule when interval changes
                """
                rq_scheduler = django_rq.get_scheduler('default')
                jobs = rq_scheduler.get_jobs()

                functions = defaultdict(lambda: list())
                map(lambda x: functions[x.func].extend([x.meta.get('interval'), x.timeout,]), jobs)

                if not timeout:
                    default_timeout = global_settings.RQ_QUEUES.get('DEFAULT_TIMEOUT')
                    if default_timeout:
                        timeout = default_timeout
                    else:
                        timeout = 360

                if f not in functions or interval not in functions[f] or functions[f][1] != timeout or len(functions[f]) > 2:
                    logger.info('Rescheduling job {} with interval: {}s'.format(f.func_name, interval))
                    # clear all scheduled jobs for this function
                    map(rq_scheduler.cancel, filter(lambda x: x.func==f, jobs))

                    # schedule with new interval and timeout
                    return rq_scheduler.schedule(datetime.datetime.now(), f, interval=interval, timeout=timeout)
                else:
                    logger.info('Job already scheduled every {}s: {}'.format(interval, f.func_name))
            f.enqueue_at = enqueue_at
            f.delay = delay
            f.schedule_once = schedule_once
            return f

    def job(func_or_queue, connection=None, *args, **kwargs):
        """
        The same as RQ's job decorator, but it works automatically works out
        the ``connection`` argument from RQ_QUEUES.

        And also, it allows simplified ``@job`` syntax to put job into
        default queue.
        """
        if callable(func_or_queue):
            func = func_or_queue
            queue = 'default'
        else:
            func = None
            queue = func_or_queue

        try:
            from django.utils import six
            string_type = six.string_types
        except ImportError:
            # for django lt v1.5 and python 2
            string_type = basestring

        if isinstance(queue, string_type):
            try:
                queue = get_queue(queue)
                if connection is None:
                    connection = queue.connection
            except KeyError:
                pass

        decorator = _rq_job(queue, connection=connection, *args, **kwargs)
        if func:
            return decorator(func)
        return decorator
    return job(func_or_queue, connection, *args, **kwargs)
